@@ -668,6 +668,10 @@ class PathMapper:
 
         minimal: dict[str, str] = {}
 
+        packaging = {
+            'integration': 'packaging/',
+        }
+
         # Early classification that needs to occur before common classification belongs here.
 
         if path.startswith('test/units/compat/'):
@@ -749,6 +753,9 @@ class PathMapper:
             return minimal
 
         if path.startswith('packaging/'):
+            if path.startswith('packaging/pep517_backend/'):
+                return packaging
+
             return minimal
 
         if path.startswith('test/ansible_test/'):
@@ -836,16 +843,17 @@ class PathMapper:
                 return minimal
 
             if path in (
-                    'setup.py',
+                'MANIFEST.in',
+                'pyproject.toml',
+                'requirements.txt',
+                'setup.cfg',
+                'setup.py',
             ):
-                return all_tests(self.args)  # broad impact, run all tests
+                return packaging
 
             if ext in (
-                '.in',
                 '.md',
                 '.rst',
-                '.toml',
-                '.txt',
             ):
                 return minimal
 
